require 'gtk2'

# If there is text in the GtkEntry, set the page as complete.
# Otherwise, stop the user from progressing the next page.
def entry_changed (entry, assistant)
  text = entry.text
  num = assistant.current_page
  page = assistant.get_nth_page(num)
  assistant.set_page_complete(page, (text.size > 0))
end

# If the check button is toggled, set the page as complete.
# Otherwise, stop the user from progressing the next page.
def button_toggled(toggle, assistant)
  num = assistant.current_page
  page = assistant.get_nth_page(num)
  assistant.set_page_complete(page, toggle.active?)
end

# Fill up the progress bar, 10% every second when the button
# is clicked. Then, set the page as complete when the progress
# bar is filled.
def button_clicked (button, assistant, progress)
  percent = 0.0
  button.sensitive = false
  num = assistant.current_page
  page = assistant.get_nth_page(num)
  while(percent <= 100.0)
    progress.text = "%.0f%% Complete" % percent
    progress.fraction = percent / 100.0
    while (Gtk.events_pending?)
      Gtk.main_iteration
    end
    sleep 0.1
    percent += 5.0
  end
  assistant.set_page_complete(page, true)
end

# This function is where you would apply the changes and
# destroy the assistant.
def assistant_close (assistant)
  puts "You would apply your changes now!\n"
  assistant.destroy
end

class PageInfo
  attr_accessor :widget, :index, :title, :type, :complete
  def initialize(w, i, ti, ty, c)
    @widget, @index, @title, @type, @complete = w, i, ti, ty, c
  end
end

page = [
  PageInfo.new(nil, -1, "Introduction",            Gtk::Assistant::PAGE_INTRO,    true),
  PageInfo.new(nil, -1, "",                        Gtk::Assistant::PAGE_CONTENT,  false),
  PageInfo.new(nil, -1, "Click the Check Button",  Gtk::Assistant::PAGE_CONTENT,  false),
  PageInfo.new(nil, -1, "Click the Button",        Gtk::Assistant::PAGE_PROGRESS, false),
  PageInfo.new(nil, -1, "Confirmation",            Gtk::Assistant::PAGE_CONFIRM,  true)
]

# Create a new assistant widget with no pages. */
assistant = Gtk::Assistant.new
assistant.set_size_request(350, 200)   # 450, 300
assistant.signal_connect('destroy') { Gtk.main_quit }

page[0].widget = Gtk::Label.new(
       "This is an example of a GtkAssistant. By\n" +
       "clicking the forward button, you can continue\n" +
       "to the next section!"
      )
page[1].widget = Gtk::HBox.new(FALSE, 5)
page[2].widget = Gtk::CheckButton.new("Click Me To Continue!")
page[3].widget = Gtk::Alignment.new(0.5, 0.5, 0.0, 0.0)
page[4].widget = Gtk::Label.new(
        "Text has been entered in the label and the\n" +
        "combo box is clicked. If you are done, then\n" +
        "it is time to leave!"
      )
# Create the necessary widgets for the second page.
label = Gtk::Label.new("Your Name: ")
entry = Gtk::Entry.new
page[1].widget.pack_start(label, false, false, 5)
page[1].widget.pack_start(entry, false, false, 5)

# Create the necessary widgets for the fourth page.
# Then, attach the progress bar to the GtkAlignment
# widget for later access.
button = Gtk::Button.new("Click me!")
progress = Gtk::ProgressBar.new
hbox = Gtk::HBox.new(false, 5)
hbox.pack_start(progress, true, false, 5)
hbox.pack_start(button,   false, false, 5)
page[3].widget.add(hbox)

# Add five pages to the Gtk::Assistant dialog.
5.times do |i|
  page[i].index = assistant.append_page(page[i].widget) 
  assistant.set_page_title(page[i].widget, page[i].title)
  assistant.set_page_type(page[i].widget, page[i].type)

  # Set the introduction and conclusion pages as complete so
  # they can be incremented or closed.
  assistant.set_page_complete(page[i].widget, page[i].complete)
end

# Update whether pages 2 through 4 are complete based upon
# whether there is text in the GtkEntry, the check button is
# active, or the progress bar is completely filled.

entry.signal_connect(         'changed') { |w| entry_changed(w, assistant) }
page[2].widget.signal_connect('toggled') { |w| button_toggled(w, assistant) }
button.signal_connect(        'clicked') { |w| button_clicked(w, assistant, progress) }
assistant.signal_connect(     'cancel')  { assistant.destroy }
assistant.signal_connect(     'close')   { |w| assistant_close(w) }

assistant.show_all
Gtk.main