#!/usr/bin/ruby

require 'gtk2'
require "rexml/document"

# constants for localization

LANGUAGE = "de"
LOCSTRINGS = {
	"de" => {
		"screen_1_title" => "Installation",
		"screen_1" => "Installieren Sie COMPUTERBILD Sicher surfen 2009 auf einen USB-Speicherstift. Stecken Sie dazu jetzt den Stift ein. Andere eventuell angeschlossene USB-Laufwerke müssen Sie entfernen.",
		"screen_2_title" => "Auswahl des Installationsziels",
		"screen_2" => "Auf welches USB-Laufwerk wollen Sie COMPUTERBILD Sicher surfen 2009 installieren?",
		"screen_2_size" => "Soviel Speicherplatz wird nach der Installation auf Ihrem Speicherstift belegt sein:",
		"screen_2_expl" => "In diesem Bereich liegt auch Ihr Datentresor. Mit  Windows können Sie darauf nicht zugreifen. Falls Sie viele vertrauliche Daten auf dem Stift speichern, können Sie den dafür reservierten Bereich größer wählen.",
		"screen_3" => "Mit dem folgenden Passwort können Sie künftig beim Start von COMPUTERBILD Sicher surfen auf Ihren Datentresor zugreifen:",
		"screen_3_user" => "Hier können Sie ein zusätzliches Passwort eingeben, das beim Zugriff auf Laufwerke abgefragt wird. Falls Sie die Felder leer lassen, wird automatisch das oben vergebene verwendet. Sie erhöhen die Sicherheit, wenn Sie verschiedene Passwörter vergeben."
	},
	"en" => {
		"screen_1_title" => "Install CB Safer Surf",
		"screen_1" => "Install COMPUTERBILD Safer Surf 2009 on a USB thumb drive. Please connect the drive now and revome other connected USB drives.",
		"screen_2_title" => "Installation target",
		"screen_2" => "Please choose the installation target:",
		"screen_2_size" => "Choose the space occupied by CB Safer Surf:",
		"screen_2_expl" => "This area also includes your encrypted container. It is not accessible from Windows. If you intend to store many files you might choose a larger value.",
		"screen_3" => "Choose a password that is used to unlock the encrypted container:",
		"screen_3_user" => "You can increase the security by choosing a diffent password for access to drives. If you do not fill out these fields, the password chosen above will be used."
	}
}

# define some common functions

class UsbDrive
	@device
	@vendor
	@product
	@size
	@partitions
	def initialize(dev)
		@device = dev
		@partitions = Hash.new
		@size = 0
		@vendor = "?"
		@product = "?"
	end
	attr_accessor :device, :vendor, :product, :partitions, :size
end

class UsbDrivePartition
	@device
	@size
	@version
	@mounted
	@mountpoint
	def initialize(dev)
		@size = 0
		@device = dev
		@mounted = false
		# now find mounted partitions
		IO.popen("mount") { |i|
			while i.gets
				tokens = $_.split
				if tokens[0] == dev
					@mounted = true
					@mountpoint = tokens[2]
				end
			end
		}
	end
	attr_accessor :device, :size, :version, :mounted, :mountpoint
end

def get_usbdrives_lshw
	all_drives = Hash.new
	xstring = ""
	begin
		IO.popen("lshw-static -xml") { |ioin|
			while ioin.gets
				xstring = xstring + $_
			end
		}
		xdoc = REXML::Document.new(xstring)
	rescue
		xdoc = REXML::Document.new("<empty><nix></nix></empty>")
	end
	# xdoc = REXML::Document.new(File.new("/tmp/lshw_sf2.xml"))
	xdoc.root.elements.each("//node[@class='storage']") { |x|
		begin
			businfo = "unknown"
			begin
				businfo = x.elements["businfo"].text.to_s
			rescue
			end
			x.elements.each("node[@class='disk']") { |element|
					if element.elements["logicalname"].text =~ /^\/dev\/sd/ && businfo =~ /^usb/
						# puts element.elements["logicalname"].text
						udrive = UsbDrive.new(element.elements["logicalname"].text)
						begin
							udrive.vendor = x.elements["vendor"].text
							udrive.product = x.elements["product"].text
						rescue
						end
						udrive.size = element.elements["size"].text.to_i
						all_parts = Hash.new
						element.elements.each("node[@class='volume']") { |vol|
							if vol.attributes["id"] =~ /^volume/  && vol.elements["description"].text =~ /fat/i
								begin
									npart = UsbDrivePartition.new(vol.elements["logicalname"].text)
									# puts vol.elements["logicalname"].text
									npart.size = vol.elements["capacity"].text.to_i
									puts vol.elements["capacity"].text.to_i
									npart.version = ""
									all_parts[vol.elements["logicalname"].text] = npart
								rescue
								end
							end
						}
						udrive.partitions = all_parts
						all_drives[element.elements["logicalname"].text] = udrive
					end
			}
			x.elements.each("node[@class='volume']") { |element|
				if element.elements["logicalname"].text =~ /^\/dev\/sd/ && businfo =~ /^usb/
					udrive = UsbDrive.new(element.elements["logicalname"].text)
					begin
						udrive.vendor = x.elements["vendor"].text
						udrive.product = x.elements["product"].text
					rescue
					end
				end
				udrive.size = element.elements["size"].text.to_i
				all_drives[element.elements["logicalname"].text] = udrive unless 
					all_drives.has_key?( element.elements["logicalname"].text )
			}
		rescue
		end
	}
	return all_drives
end

def create_drivelist(combobox, rows, usbdrives)
	(rows - 1).downto(0) { |i| 
		puts "removing " + i.to_s
		combobox.remove_text(i)
	}
	comborows = 0
	usable_drives = []
	usbdrives.each { |key, val| 
		if val.size > 900_000_000
			puts "create drivelist entry " + key
			# Use Gigabytes if the drive is 8GB or larger 
			if val.size > 7_900_000_000
				sizestr = ((val.size.to_f / 1073741824 ) + 0.5).to_i.to_s + "GB"
			else
				sizestr = (val.size.to_f / 1048576 ).to_i.to_s + "MB"
			end
			combobox.append_text( val.vendor + " " + val.product + " " + sizestr )
			usable_drives.push(val)
		end
	}
	combobox.active = 0
	return usable_drives
end

def adjust_size_spinner(spinbox, selected_item, drivelist)
	size = 0
	megsize = 0
	begin
		size = drivelist[selected_item].size
		megsize = size / 1048576
		puts "selected size: " + size.to_s
	rescue
		spinbox.sensitive = false
		return
	end
	# Allow partitioning only on drives 2GB and up
	if size > 1_900_000_000
		maxsize = megsize - 512
		maxsize = 4600 if maxsize > 4600 
		spinbox.adjustment = Gtk::Adjustment.new(1024, 1024, maxsize, 64, 1.0, 1.0)
		spinbox.value = 1024
		spinbox.sensitive = true
	else
		spinbox.adjustment = Gtk::Adjustment.new(megsize, megsize, megsize, 64, 1.0, 1.0)
		spinbox.value = megsize
		spinbox.sensitive = false
	end
	
end

def password_compare(cont_pw, cont_conf, user_pw, user_conf)
	if user_pw.size == 0 && user_conf.size == 0
		return cont_pw == cont_conf && check_pw_policy(cont_pw)
	else
		return cont_pw == cont_conf && user_pw == user_conf && check_pw_policy(cont_pw) && check_pw_policy(user_pw)
	end
	return false
end

def check_pw_policy(password) 
	return password.size > 9 && password =~ /[A-Z]/ && password =~ /[a-z]/ && password =~ /[0-9]/ 
end

def create_pwhash(pw)
	puts("== Replace password hash ==")
	# puts pw
	system(	"echo -n '" + pw.gsub("'", "\\\\" + '\'') + "' | openssl passwd -1 -stdin > /etc/lesslinux/surfer.hash") 
	system( "/etc/rc.d/0040-roothash.sh start")
end

def run_partitioning(usable_drives, target, syssize)
	puts("== Change partitioning ==")
	selected_drive = usable_drives[target]
	device =  selected_drive.device
	# drive size is in bytes!
	drive_size = selected_drive.size
	# now delete all existing partitions
	# ... the soft way -- does not always work
	# 7.downto(0) { |p| system("parted -s " + device + " rm " + p.to_s) }
	# ... the hard way
	system("dd if=/dev/zero bs=1M count=1 of=" + device)
	system("parted -s " + device + " mklabel msdos")
	system("mdev -s")
	# all further calculations in kilobytes!
	part_two_end = (drive_size.to_i / 1024) - 4096
	part_two_start = part_two_end - syssize.to_i * 1024
	part_one_end = part_two_start - 2048
	part_one_start = 63
	# if device is too small, just create one partition
	# else create two partitions
	if drive_size > 1_900_000_000
		system("parted -s " + device + " unit kB mkpartfs primary fat32 " + part_one_start.to_s + " " + part_one_end.to_s)
		system("parted -s " + device + " unit kB mkpartfs primary fat32 " + part_two_start.to_s + " " + part_two_end.to_s)
		system("parted -s " + device + " set 2 boot on ")
		system("mdev -s")
		return 2
	else
		system("parted -s " + device + " unit kB mkpartfs primary fat32 " + part_one_start.to_s + " " + part_two_end.to_s)
		system("parted -s " + device + " set 1 boot on ")
		system("mdev -s")
		return 1
	end
end

def write_bootloader(target_stick, part)
	puts("== Write bootloader ==")
	system("mdev -s")
	system("cat /usr/share/syslinux/mbr.bin > " + target_stick)
	system("mkdir /lesslinux/install_target")
	# system("mount -t ext2 " + target_stick + part.to_s + " /lesslinux/install_target")
	system("syslinux " + target_stick + part.to_s)
	# system("umount /lesslinux/install_target")
	# system("umount " + target_stick + part.to_s)
end

def run_containercopy(target_stick, part)
	puts("== Copy containers ==")
	system("mdev -s")
	system("mount -t vfat " + target_stick + part.to_s + " /lesslinux/install_target")
	# find out if install source is still mounted
	unless system("mount | awk '{print $1}' | grep ` cat /var/run/lesslinux/install_source ` ")
		system("mount ` cat /var/run/lesslinux/install_source ` /lesslinux/cdrom")
	end
	[ "syslinux.cfg", "boot", "lesslinux", "GPL.txt" ].each { |e|
		system("/bin/tar -C /lesslinux/cdrom -cf - " + e + " | /bin/tar -C  /lesslinux/install_target -xvf - ")
	}
	system("cp /etc/lesslinux/updater/version.txt /lesslinux/install_target")
	system("umount /lesslinux/install_target")
	system("umount /lesslinux/cdrom")
	system("umount " + target_stick + part.to_s)
end

def copy_autorun(target_stick)
	puts("== Copy autorun.inf ==")
	system("mount -t vfat " + target_stick + "1 /lesslinux/install_target")
	# find out if install source is still mounted
	unless system("mount | awk '{print $1}' | grep ` cat /var/run/lesslinux/install_source ` ")
		system("mount ` cat /var/run/lesslinux/install_source ` /lesslinux/cdrom")
	end
	system("cp /lesslinux/cdrom/cbicon.ico /lesslinux/install_target")
	system("cp /lesslinux/cdrom/autorun.usb /lesslinux/install_target/autorun.inf")
	system("umount /lesslinux/cdrom")
	system("umount /lesslinux/install_target")
end

def run_create_crypt(target_stick, part, syssize, pw)
	puts("== Create crypt container ==")
	system("/sbin/modprobe -v dm-crypt")
	system("/sbin/modprobe -v sha256")
	system("mdev -s")
	containersize = syssize - 512
	system("mount -t vfat " + target_stick + part.to_s + " /lesslinux/install_target")
	system("dd if=/dev/zero bs=1M count=1 seek=" + (containersize - 1).to_i.to_s + " of='/lesslinux/install_target/home.llc' " )
	# attach loop device to the container
	system("mdev -s")
	free_loop = "/dev/loop9999" # nil
	IO.popen("losetup -f") { |i| 
		while i.gets
			free_loop = $_.to_s.strip
		end
	}
	system("losetup " + free_loop + " '/lesslinux/install_target/home.llc' ")
	system("mdev -s")
	system("dd if=/dev/urandom of=" + free_loop + " bs=1M count=16")
	system("echo '" + pw.gsub("'", "\\\\" + '\'') + "' | cryptsetup luksFormat -c aes-cbc-essiv:sha256 -s 256 -q " + free_loop ) 
	# open the container
	system("echo '" + pw.gsub("'", "\\\\" + '\'') + "' | cryptsetup luksOpen -q " + free_loop + " lesslinux_crypt" ) 
	system("mdev -s")
	system("mkfs.ext2 /dev/mapper/lesslinux_crypt")
	system("mkdir  /lesslinux/tmp_home")
	system("mount -t ext2 /dev/mapper/lesslinux_crypt /lesslinux/tmp_home")
	system("rsync -avHP /home/ /lesslinux/tmp_home/")
	system("mkdir -m 0755 -p /lesslinux/tmp_home/.overlay/etc")
	system("rsync -avHP /etc/shadow /lesslinux/tmp_home/.overlay/etc")
	system("rsync -avHP /etc/passwd /lesslinux/tmp_home/.overlay/etc")
	system("umount /lesslinux/tmp_home/")
	system("cryptsetup luksClose /dev/mapper/lesslinux_crypt")
	system("losetup -d " + free_loop)
	system("umount /lesslinux/install_target")
	# system("sleep 5")
end
	
def run_installation (parent, usable_drives, target, syssize, contpass, userpass)
	puts "Installation target: " + usable_drives[target].device
	puts "Syssize: " + syssize.to_i.to_s
	align = Gtk::Alignment.new(0.5, 0.5, 1.0, 0.0)
	pbar = Gtk::ProgressBar.new
	align.add(pbar)
	dialog = Gtk::Dialog.new(
		"Installationsfortschritt",
		parent,
		Gtk::Dialog::MODAL,
		[ Gtk::Stock::OK, Gtk::Dialog::RESPONSE_OK ]
	)
	dialog.vbox.add(align)
	msg = Gtk::Label.new("Die Installation läuft. Bitte haben Sie etwas Geduld!")
	msg.wrap = true
	dialog.vbox.add(msg)
	dialog.set_size_request(450, 150)
	dialog.deletable = false
	dialog.show_all
	show_progress(pbar, dialog, msg, usable_drives, target, syssize, contpass, userpass)
	dialog.signal_connect('response') { 
		puts "trigger reboot here..."
		trigger_reboot
		Gtk.main_quit 
	}
end

def show_progress (progress, dialog, msg, usable_drives, target, syssize, contpass, userpass )
	userpass = contpass if userpass.size == 0
	
	percent = 0.0
	dialog.set_response_sensitive(Gtk::Dialog::RESPONSE_OK, false)
	# button.sensitive = false;

	progress.text = "Start der Installation: Erzeuge Passwort"
	progress.fraction = 0.02
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	create_pwhash(userpass)
	
	progress.text = "Passwort erzeugt - Partitionierung"
	progress.fraction = 0.07
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	target_partition = run_partitioning(usable_drives, target, syssize)
	
	progress.text = "Partitionierung abgeschlossen - schreibe Bootloader"
	progress.fraction = 0.15
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	write_bootloader(usable_drives[target].device, target_partition)
	
	progress.text = "Bootloader geschrieben - kopiere System"
	progress.fraction = 0.30
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	run_containercopy(usable_drives[target].device, target_partition)
	
	progress.text = "System kopiert - erstelle Datentresor"
	progress.fraction = 0.70
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	run_create_crypt(usable_drives[target].device, target_partition, syssize, contpass)
	copy_autorun(usable_drives[target].device)
	
	progress.text = "Installation abgeschlossen"
	progress.fraction = 1.00
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	msg.text = "Die Installation ist abgeschlossen, klicken Sie auf OK, um den Computer neu zu starten."
	dialog.set_response_sensitive(Gtk::Dialog::RESPONSE_OK, true)
end

def trigger_reboot
	system("/static/sbin/reboot")
end

assi = Gtk::Assistant.new
assi.deletable = false

usbdrives = nil
usable_drives = []

# Erstes Fenster, Start hat immer ID 0

desclabel = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_1"])
desclabel.wrap = true
dtid = assi.append_page(desclabel)
assi.set_page_title(desclabel, LOCSTRINGS[LANGUAGE]["screen_1_title"])
assi.set_page_type(desclabel, Gtk::Assistant::PAGE_INTRO)
assi.set_page_complete(desclabel, true)

# Zweites Fenster, Auswahl des Laufwerkes, Partitionierung

choicelabel = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_2"])
choicebox = Gtk::ComboBox.new
partlabel = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_2_size"])
partexpl = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_2_expl"])
partsize = Gtk::SpinButton.new(1024, 9999, 10)
choicebox.signal_connect('changed') {
	adjust_size_spinner(partsize, choicebox.active, usable_drives)
}
choice_vbox = Gtk::VBox.new(false, 5) 
choice_vbox.pack_start(choicelabel, false, true, 5) 
choice_vbox.pack_start(choicebox, false, true, 5) 
choice_vbox.pack_start(partlabel, false, true, 5) 
choice_vbox.pack_start(partsize, false, true, 5) 
choice_vbox.pack_start(partexpl, false, true, 5) 
choicelabel.wrap = true
choicelabel.width_request = 450
choicelabel.xalign = 0.1
partlabel.wrap = true
partlabel.width_request = 450
partlabel.xalign = 0.1
partexpl.wrap = true
partexpl.xalign = 0.1
partexpl.width_request = 450
choiceid = assi.append_page(choice_vbox)
assi.set_page_title(choice_vbox, LOCSTRINGS[LANGUAGE]["screen_2_title"])
assi.set_page_type(choice_vbox, Gtk::Assistant::PAGE_CONTENT)
assi.set_page_complete(choice_vbox, false)

# Drittes Fenster: Passwörter

contlabel = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_3"])
userlabel = Gtk::Label.new(LOCSTRINGS[LANGUAGE]["screen_3_user"])
pwdetlabel = Gtk::Label.new("Das Passwort muss wenigstens zehn Zeichen lang sein und mindestens einen Klein-, einen Großbuchstaben sowie eine Zahl enthalten.")
contlabel.wrap = true
userlabel.wrap = true
pwdetlabel.wrap = true
contlabel.width_request = 550
userlabel.width_request = 550
pwdetlabel.width_request = 550
contlabel.xalign = 0.1
userlabel.xalign = 0.1
pwdetlabel.xalign = 0.1

d_table = Gtk::Table.new(2,    2,    false)
dt_options = Gtk::EXPAND|Gtk::FILL
d_pass = Gtk::Entry.new
d_pass.visibility = false
c_pass = Gtk::Entry.new
c_pass.visibility = false
d_table.attach(d_pass,  1,  2,  0,  1, nil, nil, 5, 2)
d_table.attach(c_pass,  1,  2,  1,  2, nil, nil, 5, 2)
p_label = Gtk::Label.new("Passwort:")
p_label.xalign = 0
p_label.width_request = 100
c_label = Gtk::Label.new("Bestätigung:")
c_label.xalign = 0
c_label.width_request = 100
d_table.attach(p_label,  0,  1,  0,  1, dt_options, nil, 5, 2)
d_table.attach(c_label,  0,  1,  1,  2, dt_options, nil, 5, 2)

u_table = Gtk::Table.new(2,    2,    false)
u_pass = Gtk::Entry.new
u_pass.visibility = false
uc_pass = Gtk::Entry.new
uc_pass.visibility = false
u_table.attach(u_pass,  1,  2,  0,  1, nil, nil, 5, 2)
u_table.attach(uc_pass,  1,  2,  1,  2, nil, nil, 5, 2)
u_label = Gtk::Label.new("Passwort:")
u_label.xalign = 0
u_label.width_request = 100
uc_label = Gtk::Label.new("Bestätigung:")
uc_label.xalign = 0
uc_label.width_request = 100
u_table.attach(u_label,  0,  1,  0,  1, dt_options, nil, 5, 2)
u_table.attach(uc_label,  0,  1,  1,  2, dt_options, nil, 5, 2)

contvbox = Gtk::VBox.new(false, 5)
contvbox.pack_start(contlabel, false, true, 5) 
contvbox.pack_start(d_table, false, true, 5) 
contvbox.pack_start(userlabel, false, true, 5)
contvbox.pack_start(u_table, false, true, 5)
contvbox.pack_start(pwdetlabel, false, true, 5)
contid = assi.append_page(contvbox)
assi.set_page_title(contvbox, "Passwörter")
assi.set_page_type(contvbox, Gtk::Assistant::PAGE_CONTENT)
assi.set_page_complete(contvbox, false)

d_pass.signal_connect("key-release-event") { 
	assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
}
c_pass.signal_connect("key-release-event") { 
	assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
}
u_pass.signal_connect("key-release-event") { 
	assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
}
uc_pass.signal_connect("key-release-event") { 
	assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
}

# Letztes Label: Fertig, Neustart

endlabel = Gtk::Label.new("Klicken Sie auf \"Anwenden\" um die Installation mit den vorgenommenen Einstellungen zu starten. Alle Daten auf dem ausgewählten Laufwerk werden dabei gelöscht.")
endlabel.wrap = true
endid = assi.append_page(endlabel)
assi.set_page_title(endlabel, "Bereit zur Installation")
assi.set_page_type(endlabel, Gtk::Assistant::PAGE_CONFIRM)
assi.set_page_complete(endlabel, true)

assi.set_title  "Installation COMPUTERBILD Sicher surfen 2009"
assi.border_width = 10
assi.set_size_request(600, 410)

assi.set_forward_page_func{|curr|
	puts curr
	if curr == 0
		# when switching from first to second screen, read drivelist
		puts "current screen 0, rereading drivelist"
		usbdrives = get_usbdrives_lshw
		puts usbdrives
		usable_drives = create_drivelist(choicebox, usable_drives.size, usbdrives)
		assi.set_page_complete(choice_vbox, usable_drives.size > 0)
		1
	else
		curr + 1
	end
}

assi.signal_connect('close')   { |w|
	run_installation(w, usable_drives, choicebox.active, partsize.value, c_pass.text, uc_pass.text)
}

assi.signal_connect('cancel')   { |w|
	puts "cleanup and reboot here"
	trigger_reboot
	Gtk.main_quit 
}

assi.signal_connect('delete_event') { 
	puts "cleanup and reboot here"
	trigger_reboot
	Gtk.main_quit 
}

assi.show_all
Gtk.main

