#!/usr/bin/ruby

require 'gtk2'
require "rexml/document"

# FIXME: Put these functions to a separate file

def password_compare(user_pw, user_conf, root_pw, root_conf)
	if root_pw.size == 0 && root_conf.size == 0
		return user_pw == user_conf && check_pw_policy(user_pw)
	elsif user_pw.size == 0 && user_conf.size == 0
		return root_pw == root_conf && check_pw_policy(root_pw)
	else
		return user_pw == user_conf && root_pw == root_conf && check_pw_policy(user_pw) && check_pw_policy(root_pw)
	end
	return false
end

def check_pw_policy(password) 
	return password.size > 9 && password =~ /[A-Z]/ && password =~ /[a-z]/ && password =~ /[0-9]/ 
end

def create_pw_hash(user, pw)
	puts("== Replace password hash ==")
	# puts pw
	system(	"echo -n '" + pw.gsub("'", "\\\\" + '\'') + "' | openssl passwd -1 -stdin > /etc/lesslinux/" + user + ".hash") 
	system( "/etc/rc.d/0040-roothash.sh start")
end

# try to find the partition with home.llc

def find_home_container
	default_container = "/lesslinux/cdrom/home.llc"
	if File.exist?(default_container)
		system("mount -o remount,rw /lesslinux/cdrom")
		return default_container
	end
	volumes = Array.new
	lcount = 0
	# Find all partitions identified by lshw
	IO.popen("lshw-static -quiet -short -class volume") { |i|
		while i.gets
			cols = $_.split(/\s{2,}/)
			if lcount > 1 && cols[3].to_s =~ /FAT/i
				volumes.push(cols[1])
			end
			lcount += 1
		end
	}
	system("mkdir /lesslinux/cryptpart")
	volumes.each { |i|
		system("mount -o ro " + i + " /lesslinux/cryptpart")
		default_container = "/lesslinux/cryptpart/home.llc"
		if File.exist?(default_container)
			system("mount -o remount,rw /lesslinux/cryptpart")
			return default_container
		end
		system("umount /lesslinux/cryptpart")
	}
	return nil
end

def save_xorg_settings( driver_combo, resolutions, res_combo, monitor_modes, mon_combo )
	outfile = File.new("/var/run/lesslinux/xorg_vars", "w")
	# outfile = File.new("/tmp/xorg_vars", "w")
	if driver_combo.active == 0
		outfile.write("driver=xorg\n")
	else
		outfile.write("driver=xvesa\n")
	end
	res = resolutions[res_combo.active]
	outfile.write("vesapref=" + res + "x24\n")
	outfile.write("screen=Screen_" + monitor_modes[mon_combo.active] + "_" + res + "\n")
	outfile.close
end

def mount_crypt_container(container, pw)
	puts("== Mount crypt container ==")
	system("/sbin/modprobe -v dm-crypt")
	system("/sbin/modprobe -v sha256")
	system("mdev -s")
	free_loop = "/dev/loop9999" # nil
	IO.popen("losetup -f") { |i| 
		while i.gets
			free_loop = $_.to_s.strip
		end
	}
	system("losetup " + free_loop.to_s + " " + container.to_s )
	unless system("echo '" + pw.gsub("'", "\\\\" + '\'') + "' | cryptsetup luksOpen -q " + free_loop + " lesslinux_crypt" )
		system( "cryptsetup luksClose lesslinux_crypt" )
		system( "losetup -d " + free_loop )
		system( "umount /lesslinux/cryptpart" )
		return false, "Fehler: Falsches Passwort oder beschädigter Datentresor! Der Start wird ohne Datentresor fortgesetzt."
	end
	unless system("mount -t ext2 -o nodev,nosuid,noexec /dev/mapper/lesslinux_crypt /home")
		# try to repair fs, preen should be safe enough
		system("/sbin/e2fsck -p /dev/mapper/lesslinux_crypt") 
		unless system("mount -t ext2 -o nodev,nosuid,noexec /dev/mapper/lesslinux_crypt /home")
			system("cryptsetup luksClose lesslinux_crypt")
			system("losetup -d " + free_loop)
			system("umount /lesslinux/cryptpart")
			return false, "Zugriff auf Datentresor nicht möglich: Korrektes Passwort, aber beschädigtes Dateisystem!"
		end
	else
		system("rsync -avHP /home/.overlay/ /")
		return true, "Zugriff auf Datentresor erfolgreich!"
	end
	return false, "This should not happen!"
end

def show_error_dialog(message)
	dialog = Gtk::MessageDialog.new($main_application_window, 
			Gtk::Dialog::MODAL,
			Gtk::MessageDialog::ERROR,
			Gtk::MessageDialog::BUTTONS_CLOSE,
			message)
	dialog.run
	dialog.destroy
end

def trigger_reboot
	system("/static/sbin/reboot")
end

# build the GUI
# first step is mounting the encrypted container

crypt_container = find_home_container

assi = Gtk::Assistant.new
assi.deletable = false

# first window, create vbox for entry of user and root password

contlabel = Gtk::Label.new("Hier können Sie ein Passwort für den Zugriff auf Laufwerke vergeben. Wenn Sie kein Passwort vergeben ist kein Zugriff auf Laufwerke möglich.")
userlabel = Gtk::Label.new("Ein Administratorpasswort wird nur in seltenen Fällen, beispielsweise für Support-Aufgaben benötigt.")
pwdetlabel = Gtk::Label.new("Das vergebene Passwort muss wenigstens zehn Zeichen lang sein und mindestens einen Klein-, einen Großbuchstaben sowie eine Zahl enthalten.")
contlabel.wrap = true
userlabel.wrap = true
pwdetlabel.wrap = true
contlabel.width_request = 550
userlabel.width_request = 550
pwdetlabel.width_request = 550
contlabel.xalign = 0.1
userlabel.xalign = 0.1
pwdetlabel.xalign = 0.1

d_table = Gtk::Table.new(2,    2,    false)
dt_options = Gtk::EXPAND|Gtk::FILL
d_pass = Gtk::Entry.new
d_pass.visibility = false
c_pass = Gtk::Entry.new
c_pass.visibility = false
d_table.attach(d_pass,  1,  2,  0,  1, nil, nil, 5, 2)
d_table.attach(c_pass,  1,  2,  1,  2, nil, nil, 5, 2)
p_label = Gtk::Label.new("Passwort:")
p_label.xalign = 0
p_label.width_request = 100
c_label = Gtk::Label.new("Bestätigung:")
c_label.xalign = 0
c_label.width_request = 100
d_table.attach(p_label,  0,  1,  0,  1, dt_options, nil, 5, 2)
d_table.attach(c_label,  0,  1,  1,  2, dt_options, nil, 5, 2)

u_table = Gtk::Table.new(2,    2,    false)
u_pass = Gtk::Entry.new
u_pass.visibility = false
uc_pass = Gtk::Entry.new
uc_pass.visibility = false
u_table.attach(u_pass,  1,  2,  0,  1, nil, nil, 5, 2)
u_table.attach(uc_pass,  1,  2,  1,  2, nil, nil, 5, 2)
u_label = Gtk::Label.new("Passwort:")
u_label.xalign = 0
u_label.width_request = 100
uc_label = Gtk::Label.new("Bestätigung:")
uc_label.xalign = 0
uc_label.width_request = 100
u_table.attach(u_label,  0,  1,  0,  1, dt_options, nil, 5, 2)
u_table.attach(uc_label,  0,  1,  1,  2, dt_options, nil, 5, 2)

contvbox = Gtk::VBox.new(false, 5)
contvbox.pack_start(contlabel, false, true, 5) 
contvbox.pack_start(d_table, false, true, 5) 
# contvbox.pack_start(userlabel, false, true, 5)
# contvbox.pack_start(u_table, false, true, 5)
contvbox.pack_start(pwdetlabel, false, true, 5)

# first window alternatively create vbox for entry of container password

enclabel = Gtk::Label.new("Geben Sie das Passwort für Ihren Datentresor ein. Falls Sie das Eingabefeld leer lassen, startet COMPUTERBILD Sicher Surfen 2009 ohne Zugriff auf Ihre persönlichen Einstellungen.")
enclabel.wrap = true
enclabel.width_request = 550
enclabel.xalign = 0.1
enc_table = Gtk::Table.new(1,    2,    false)
enct_options = Gtk::EXPAND|Gtk::FILL
enc_pass = Gtk::Entry.new
enc_pass.visibility = false
enc_table.attach(enc_pass,  1,  2,  0,  1, nil, nil, 5, 2)
enc_ent_label = Gtk::Label.new("Passwort:")
enc_ent_label.xalign = 0
enc_ent_label.width_request = 100
enc_table.attach(enc_ent_label,  0,  1,  0,  1, enct_options, nil, 5, 2)
encvbox = Gtk::VBox.new(false, 5)
encvbox.pack_start(enclabel, false, true, 5)
encvbox.pack_start(enc_table, false, true, 5)

if crypt_container.nil?
	contid = assi.append_page(contvbox)
	assi.set_page_title(contvbox, "Passwörter")
	assi.set_page_type(contvbox, Gtk::Assistant::PAGE_CONTENT)
	assi.set_page_complete(contvbox, true) # false)
	d_pass.signal_connect("key-release-event") { 
		assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
	}
	c_pass.signal_connect("key-release-event") { 
		assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
	}
	u_pass.signal_connect("key-release-event") { 
		assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
	}
	uc_pass.signal_connect("key-release-event") { 
		assi.set_page_complete(contvbox, password_compare(d_pass.text, c_pass.text, u_pass.text, uc_pass.text))
	}
else
	contid = assi.append_page(encvbox)
	assi.set_page_title(encvbox, "Zugriff auf Datentresor")
	assi.set_page_type(encvbox, Gtk::Assistant::PAGE_CONTENT)
	assi.set_page_complete(encvbox, true) # false)
end

# Second and last page: create monitor selection

drivers = [ "Xorg (automatische Auswahl)", "VESA (langsam, funktioniert immer)" ]
resolutions = [ 
		"1920x1200",  
		"1920x1080", 
		"1680x1050", 
		"1600x1200", 
		"1440x900", 
		"1400x1050", 
		"1366x768", 
		"1280x1024", 
		"1280x800", 
		"1024x768", 
		"1024x600", 
		"800x600", 
		"800x480", 
		"640x480" ]
monitors = [ "automatisch erkennen", "Flachbildschirm/TFT (60Hz)", "Röhrenmonitor (75Hz)" ]
monitor_modes = [ "AUTO", "TFT", "CRT" ]
monitor_layout_table = Gtk::Table.new(5, 2, false)
driver_combo = Gtk::ComboBox.new
drivers.each { |i| driver_combo.append_text(i) }
res_combo = Gtk::ComboBox.new
resolutions.each { |i| res_combo.append_text(i) }
res_combo.active = 9
mon_combo = Gtk::ComboBox.new
monitors.each { |i| mon_combo.append_text(i) }
mon_combo.active = 0
button_box = Gtk::HBox.new(true, 5)
test_button = Gtk::Button.new("Test")
test_button.signal_connect("clicked") { |w|
	outfile = File.new("/var/run/lesslinux/xorg_vars.test", "w")
	# outfile = File.new("/tmp/xorg_vars", "w")
	if driver_combo.active == 0
		outfile.write("driver=xorg\n")
	else
		outfile.write("driver=xvesa\n")
	end
	res = resolutions[res_combo.active]
	outfile.write("vesapref=" + res + "\n")
	outfile.write("screen=Screen_" + monitor_modes[mon_combo.active] + "_" + res + "\n")
	outfile.close
	system("/bin/bash /usr/local/xconfgui/trigger_configtest")
}

# save_button = Gtk::Button.new("Speichern")
button_box.pack_start(test_button, true, true, 5)
# button_box.pack_start(save_button, true, true, 5)
# monitor_layout_table.attach(head, 0, 2, 0, 1 , Gtk::EXPAND, Gtk::SHRINK, 5, 15)
monitor_layout_table.attach(driver_combo,  1,  2,  1,  2, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(Gtk::Label.new("Treiber:"), 0,  1,  1,  2, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(res_combo,  1,  2,  2,  3, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(Gtk::Label.new("Auflösung:"), 0,  1,  2,  3, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(mon_combo,  1,  2,  3,  4, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(Gtk::Label.new("Monitor:"), 0,  1,  3,  4, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
monitor_layout_table.attach(button_box, 0,  2,  4,  5, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)

if File.exists?("/var/run/lesslinux/xconfgui_xorg")
	driver_combo.active = 0
else
	driver_combo.active = 1
	test_button.sensitive = false
end
if File.exists?("/var/run/lesslinux/xconfgui_skip_monitor")
	driver_combo.active = 0
	driver_combo.sensitive = false
	(resolutions.size - 1).downto(0) { |i|
		res_combo.remove_text(i)
	}
	res_combo.append_text("bereits automatisch eingestellt")
	res_combo.active = 0
	res_combo.sensitive = false
	mon_combo.sensitive = false
	test_button.sensitive = false
	
end
endid = assi.append_page(monitor_layout_table)
assi.set_page_title(monitor_layout_table, "Monitor und Grafikkarte")
assi.set_page_type(monitor_layout_table, Gtk::Assistant::PAGE_CONFIRM)
assi.set_page_complete(monitor_layout_table, true)

# common settings

assi.set_title  "Systemstart COMPUTERBILD Sicher surfen 2009"
assi.border_width = 10
assi.set_size_request(600, 410)

assi.signal_connect('close')   { |w|
	# save settings
	# user password
	create_pw_hash("surfer", d_pass.text) unless d_pass.text.size < 1
	# root password
	create_pw_hash("root", u_pass.text) unless u_pass.text.size < 1
	# save Xorg settings
	save_xorg_settings( driver_combo, resolutions, res_combo, monitor_modes, mon_combo ) unless 
		File.exists?("/var/run/lesslinux/xconfgui_skip_monitor")
	# mount the crypt container
	unless enc_pass.text.size < 1
		success, message = mount_crypt_container(crypt_container, enc_pass.text) 
		show_error_dialog(message) if success == false
	else
		system("umount /lesslinux/cryptpart")
	end
	Gtk.main_quit 
}

assi.signal_connect('cancel')   { |w|
	puts "cleanup and reboot here?"
	trigger_reboot
	Gtk.main_quit 
}

assi.signal_connect('delete_event') { 
	puts "cleanup and reboot here?"
	trigger_reboot
	Gtk.main_quit 
}

assi.show_all
Gtk.main

