#!/usr/bin/env ruby

require 'gtk2'
require "rexml/document"

# define functions to trigger

class UsbDrive
	@device
	@vendor
	@product
	@size
	@partitions
	def initialize(dev)
		@device = dev
		@partitions = Hash.new
		@size = 0
		@vendor = "?"
		@product = "?"
	end
	attr_accessor :device, :vendor, :product, :partitions, :size
end

class UsbDrivePartition
	@device
	@size
	@version
	@mounted
	@mountpoint
	def initialize(dev)
		@size = 0
		@device = dev
		@mounted = false
		# now find mounted partitions
		IO.popen("mount") { |i|
			while i.gets
				tokens = $_.split
				if tokens[0] == dev
					@mounted = true
					@mountpoint = tokens[2]
				end
			end
		}
	end
	attr_accessor :device, :size, :version, :mounted, :mountpoint
end

def get_usbdrives_lshw
	all_drives = Hash.new
	xstring = ""
	begin
		IO.popen("lshw-static -xml") { |ioin|
			while ioin.gets
				xstring = xstring + $_
			end
		}
		xdoc = REXML::Document.new(xstring)
	rescue
		xdoc = REXML::Document.new("<empty><nix></nix></empty>")
	end
	# xdoc = REXML::Document.new(File.new("/tmp/lshw2.xml"))
	xdoc.root.elements.each("//node[@class='storage']") { |x|
		begin
			businfo = "unknown"
			begin
				businfo = x.elements["businfo"].text.to_s
			rescue
			end
			x.elements.each("node[@class='disk']") { |element|
					if element.elements["logicalname"].text =~ /^\/dev\/sd/ && businfo =~ /^usb/
						# puts element.elements["logicalname"].text
						udrive = UsbDrive.new(element.elements["logicalname"].text)
						begin
							udrive.vendor = x.elements["vendor"].text
							udrive.product = x.elements["product"].text
						rescue
						end
						udrive.size = element.elements["size"].text.to_i
						all_parts = Hash.new
						element.elements.each("node[@class='volume']") { |vol|
							if vol.attributes["id"] =~ /^volume/  && vol.elements["description"].text =~ /fat/i
								begin
									npart = UsbDrivePartition.new(vol.elements["logicalname"].text)
									# puts vol.elements["logicalname"].text
									npart.size = vol.elements["capacity"].text.to_i
									puts vol.elements["capacity"].text.to_i
									npart.version = ""
									all_parts[vol.elements["logicalname"].text] = npart
								rescue
								end
							end
						}
						udrive.partitions = all_parts
						all_drives[element.elements["logicalname"].text] = udrive
					end
			}
		rescue
		end
	}
	return all_drives
end

def get_drive_details(partition)
	lcount = 0
	disk_name = nil
	disk_device = nil
	partitions = []
	IO.popen("lshw-static -quiet -short -class disk") { |i|
		while i.gets
			tokens = $_.split(/\s{2,}/)
			if lcount > 1 && partition.index(tokens[1].strip) == 0
				disk_device = tokens[1].strip
				disk_name = tokens[3].strip
			end
			lcount += 1
		end
	}
	lcount = 0
	IO.popen("lshw-static -quiet -short -class volume") { |i|
		while i.gets
			tokens = $_.split(/\s{2,}/)
			if lcount > 1 && tokens[1].strip.index(disk_device) == 0
				partitions.push( tokens[1].strip )
			end
			lcount += 1
		end
	}
	return disk_device, disk_name, partitions
end

def find_containers
	partitions = []
	mnt_parts = []
	containers = []
	lcount = 0
	# Find all partitions identified by lshw
	IO.popen("lshw-static -quiet -short -class volume") { |i|
		while i.gets
			tokens = $_.split(/\s{2,}/)
			if lcount > 1 && tokens[3].to_s =~ /FAT/i
				partinfo = [ tokens[1], tokens[3] ]
				partitions.push(partinfo)
			end
			lcount += 1
		end
	}
	# now find mounted partitions
	IO.popen("mount") { |i|
		while i.gets
			tokens = $_.split
			mountinfo = [ tokens[0], tokens[2] ]
			mnt_parts.push(mountinfo)
		end
	}
	# mount each partition and try to find containers in the root 
	# directory and the directory "lesslinux" FIXME: For now just 
	# root directory!
	partitions.each { |p|
		mountpoint = "/tmp/llcontsearch"
		mnt_parts.each { |m|
			mountpoint = m[1] if p[0] == m[0] 
		}
		if mountpoint == "/tmp/llcontsearch"
			system("mkdir " + mountpoint)
			system("mount -t vfat " + p[0] + " " + mountpoint )
		end
		Dir.foreach(mountpoint) { |c|
			if c =~ /\.llc$/i 
				containers.push( [ p[0], c ] ) 
			end
		}
		if mountpoint == "/tmp/llcontsearch"
			system("umount " + mountpoint)
		end
	}
	return partitions, containers, mnt_parts
end

def change_password(user)
	# Create the dialog
	if user == "root"
		msg = "Passwort des Administrators setzen: Bitte geben Sie das Passwort zweimal ein." 
	else
		msg = "Nutzer-Passwort setzen: Bitte geben Sie das Passwort zweimal ein."
	end
	dialog = Gtk::Dialog.new("Passwort setzen",
		$main_application_window,
		Gtk::Dialog::MODAL,
		[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT],
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_ACCEPT])
	d_table = Gtk::Table.new(4,    2,    false)
	dt_options = Gtk::EXPAND|Gtk::FILL
	d_table.attach(Gtk::Label.new(msg),  0,  2,  0,  1, nil, nil, 5, 5)
	d_pass = Gtk::Entry.new
	d_pass.visibility = false
	c_pass = Gtk::Entry.new
	c_pass.visibility = false
	d_table.attach(d_pass,  1,  2,  1,  2, nil, nil, 5, 2)
	d_table.attach(c_pass,  1,  2,  2,  3, nil, nil, 5, 2)
	p_label = Gtk::Label.new("Passwort:")
	c_label = Gtk::Label.new("Bestätigung:")
	d_table.attach(p_label,  0,  1,  1,  2, nil, nil, 5, 2)
	d_table.attach(c_label,  0,  1,  2,  3, nil, nil, 5, 2)
	i_label = Gtk::Label.new("")
	d_table.attach(i_label,  0,  2,  3,  4, nil, nil, 5, 2)

	# Ensure that the dialog box is destroyed when the user responds.
	dialog.signal_connect('response') { |x|
		puts x[1]
		if (x[1] == -2)
			dialog.destroy
		else
			result, message = pw_check(c_pass.text, d_pass.text)
			if result == false
				i_label.text = message
			else
				system(	"echo -n " + c_pass.text + 
					" | openssl passwd -1 -stdin > /etc/lesslinux/" + user + ".hash") 
				system( "/etc/rc.d/0040-roothash.sh start")
				dialog.destroy
			end
		end
		# puts x[1]
		# i_label.text = pw_check(x[1])
	}
	dialog.vbox.add(d_table)
	dialog.show_all
end

def pw_check (pass, conf)
	if [pass.size, conf.size].min < 8 
		return false, "Das Passwort muss mindestens acht Zeichen lang sein!"
	elsif pass != conf
		return false, "Die Passwörter müssen übereinstimmen!"
	else
		return true, "Die Passwörter stimmen überein."
	end
	#if x == -2
	#	return "Das Passwort muss mindestens acht Zeichen lang sein!"
	#else
	#	return "Die Passwörter müssen übereinstimmen!"
	#end
end

def mount_crypt_dialog
	msg = "Verschlüsselten Container benutzen, bitte geben Sie das Passwort ein:"
	dialog = Gtk::Dialog.new("Container benutzen",
		$main_application_window,
		Gtk::Dialog::MODAL,
		[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT],
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_ACCEPT])
	d_table = Gtk::Table.new(3,    2,    false)
	dt_options = Gtk::EXPAND|Gtk::FILL
	d_table.attach(Gtk::Label.new(msg),  0,  2,  0,  1, nil, nil, 5, 5)
	d_pass = Gtk::Entry.new
	d_pass.max_length = 65000
	d_pass.visibility = false
	d_table.attach(d_pass,  1,  2,  1,  2, nil, nil, 5, 2)
	p_label = Gtk::Label.new("Passwort:")
	d_table.attach(p_label,  0,  1,  1,  2, nil, nil, 5, 2)
	i_label = Gtk::Label.new("")
	d_table.attach(i_label,  0,  2,  2,  3, nil, nil, 5, 2)
	dialog.vbox.add(d_table)
	dialog.show_all
	password = nil
	retval = false
	if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
		password = d_pass.text
		retval = true
	end
	dialog.destroy
	return retval, password
end

def mount_crypt_container(containers, mnt_parts, radio1, radio2, radio3, combobox, passwd)
	if radio1.active?
		return false, "Kein Container ausgewählt!"
	end
	if system("mountpoint /lesslinux/cryptpart")
		return false, "Der Ordner /lesslinux/cryptpart ist bereits ein Mountpoint!"
	end
	# FIXME: identify containers on mounted partitions and change mount mode to rw!
	puts "Radio 1, state " + radio1.active?.to_s
	puts "Radio 2, state " + radio2.active?.to_s
	puts "Radio 3, state " + radio3.active?.to_s
	puts "Combobox, state " + combobox.active.to_s
	system("/sbin/modprobe -v dm-crypt")
	system("/sbin/modprobe -v sha256")
	mnt_parts = []
	# now find mounted partitions
	IO.popen("mount") { |i|
		while i.gets
			tokens = $_.split
			mountinfo = [ tokens[0], tokens[2] ]
			mnt_parts.push(mountinfo)
		end
	}
	# trying to mount the partition with the container / should not fail
	container_partition = containers[combobox.active][0]
	container_file = containers[combobox.active][1]
	
	cont_part_mountpoint = nil
	mnt_parts.each { |p|
		if p[0].strip == container_partition
			cont_part_mountpoint = p[1].strip
		end
	}
	
	system("mkdir -m 0700 /lesslinux/cryptpart")
	if cont_part_mountpoint.nil?
		cont_part_mountpoint = "/lesslinux/cryptpart" 
		unless system("mount -t vfat " + container_partition + " " + cont_part_mountpoint )
			return false, "Partition mit dem Container konnte nicht eingebunden werden!"
		end
	else
		if radio2.active?
			system("mount -o remount,rw " + cont_part_mountpoint )
		end
	end
	# trying to create loop device for container / should not fail
	free_loop = "/dev/loop999"
	system("mdev -s")
	IO.popen("losetup -f") { |i| 
		while i.gets
			free_loop = $_.to_s.strip
		end
	}
	unless system("losetup " + free_loop + " '" + cont_part_mountpoint + "/" +  container_file + "'")
		system("losetup -d " + free_loop)
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht erstellt werden!"
	else
		system("mdev -s")
	end
		
	# trying to cryptsetup container / might fail because of wrong password
	#                                / might fail because of invalid container
	unless system("echo '" + passwd.to_s + "' | " + 
		" cryptsetup luksOpen " + free_loop.to_s + " lesslinux_crypt")
		system("losetup -d " + free_loop)
		system("umount /lesslinux/cryptpart")
		return false, "Zugriff auf Container nicht möglich: Falscher Schlüssel oder beschädigter Container!"
	else
		system("mdev -s")
	end
	# trying to mount container / might fail because of defect file system
	
	if radio2.active?
		unless system("mount -t ext2 /dev/mapper/lesslinux_crypt /home")
			# try to repair fs, preen should be safe enough
			system("/sbin/e2fsck -p /dev/mapper/lesslinux_crypt") 
			unless system("mount -t ext2 -o nodev,nosuid,noexec /dev/mapper/lesslinux_crypt /home")
				system("cryptsetup luksClose lesslinux_crypt")
				system("losetup -d " + free_loop)
				system("umount /lesslinux/cryptpart")
				return false, "Zugriff auf Container nicht möglich: Unbekanntes oder beschädigtes Dateisystem!"
			end
		end
		system("rsync -avHP /home/.overlay/ /")
		return true, "Einbinden des Containers erfolgreich!"			
	elsif radio3.active?
		system("mkdir /lesslinux/crypttmp")
		unless system("mount -t ext2 /dev/mapper/lesslinux_crypt /lesslinux/crypttmp")
			system("cryptsetup luksClose lesslinux_crypt")
			system("losetup -d " + free_loop)
			system("umount /lesslinux/cryptpart")
			return false, "Zugriff auf Container nicht möglich: Unbekanntes oder beschädigtes Dateisystem!"
		end
		system("mount -t tmpfs tmpfs /home")
		system("rsync -avHP --delete /lesslinux/crypttmp/ /home/")
		system("rm -rf /home/.overlay")
		system("rsync -avHP /lesslinux/crypttmp/.overlay/ /")
		system("umount /lesslinux/crypttmp")
		system("cryptsetup luksClose lesslinux_crypt")
		system("losetup -d " + free_loop)
		system("umount /lesslinux/cryptpart")
		return true, "Einbinden des Containers und kopieren der Daten erfolgreich."
	end
	system("umount /lesslinux/crypttmp")
	system("cryptsetup luksClose lesslinux_crypt")
	system("losetup -d " + free_loop)
	system("umount /lesslinux/cryptpart")
	return false, "This should not happen!"
end

def create_container_dialog(partitions)
	partitions, containers, mnt_parts = find_containers
	msg = "Bitte bestimmen Sie Größe, Name und Partition Ihres verschlüsselten Containers:"
	dialog = Gtk::Dialog.new("Container benutzen",
		$main_application_window,
		Gtk::Dialog::MODAL,
		[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT],
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_ACCEPT])
	d_table = Gtk::Table.new(6,    2,    false)
	d_label = Gtk::Label.new(msg)
	d_label.wrap = true
	d_table.attach(d_label,  0,  2,  0,  1, nil, nil, 5, 2)
	d_name = Gtk::Entry.new
	d_name.text = "home.llc"
	d_table.attach(Gtk::Label.new("Name:"),  0,  1,  1,  2, nil, nil, 5, 2)
	d_table.attach(d_name,  1,  2,  1,  2, nil, nil, 5, 2)
	d_table.attach(Gtk::Label.new("Laufwerk:"),  0,  1,  2,  3, nil, nil, 5, 2)
	d_part = Gtk::ComboBox.new
	d_part.width_request = 360
	partitions.each { |c|
		d_part.append_text( c[0].strip + " (" + c[1].strip + ")" )
	}
	d_part.active = 0
	d_table.attach(d_part,  1,  2,  2,  3, nil, nil, 5, 2)
	d_size = Gtk::SpinButton.new(Gtk::Adjustment.new(512, 64, 4032, 32, 32, 1), 32, 0 )
	d_size.snap_to_ticks = true
	d_table.attach(Gtk::Label.new("Größe (MB):"),  0,  1,  3,  4, nil, nil, 5, 2)
	d_table.attach(d_size,  1,  2,  3,  4, nil, nil, 5, 2)
	d_pass = Gtk::Entry.new
	d_pass.visibility = false
	c_pass = Gtk::Entry.new
	c_pass.visibility = false
	d_table.attach(Gtk::Label.new("Passwort:"),  0,  1,  4,  5, nil, nil, 5, 2)
	d_table.attach(Gtk::Label.new("Bestätigung:"),  0,  1,  5,  6, nil, nil, 5, 2)
	d_table.attach(d_pass,  1,  2,  4,  5, nil, nil, 5, 2)
	d_table.attach(c_pass,  1,  2,  5,  6, nil, nil, 5, 2)
	dialog.vbox.add(d_table)
	dialog.show_all
	retval = false
	if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
		retval = true
	end
	act_part = partitions[d_part.active]
	size = d_size.value.to_i
	pass = d_pass.text
	conf = c_pass.text
	container_name = d_name.text
	dialog.destroy
	return retval, act_part[0], container_name, size, pass, conf
end

def create_crypt_container(partition, size, container_name, passphrase)
	if system("mountpoint /lesslinux/cryptpart")
		return false, "Der Ordner /lesslinux/cryptpart ist bereits ein Mountpoint!"
	end
	system("/sbin/modprobe -v dm-crypt")
	system("/sbin/modprobe -v sha256")
	# mount the partition
	system("mkdir -m 0700 /lesslinux/cryptpart")
	unless system("mount -t vfat " + partition + " /lesslinux/cryptpart")
		return false, "Mount der Partition " + partition + " fehlgeschlagen - wahrscheinlich defektes Dateisystem?"
	end
	# FIXME: check for space first
	# check for existing file
	Dir.foreach("/lesslinux/cryptpart") { |f|
		if f == container_name
			system("umount /lesslinux/cryptpart")
			return false, "Eine Datei " + container_name + " existiert bereits."
		end
	}
	# create the empty file
	unless system("dd if=/dev/zero bs=1M count=1 seek=" + (size - 1).to_s + " of='/lesslinux/cryptpart/" + container_name + "' " )
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Erstellen des Containers fehlgeschlagen - zu wenig freier Platz?"
	end
	# attach loop device to the container
	system("mdev -s")
	free_loop = nil
	IO.popen("losetup -f") { |i| 
		while i.gets
			free_loop = $_.to_s.strip
		end
	}
	if free_loop.nil? 
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Kein freies Loop-Device!"
	else
		unless system("losetup " + free_loop + " '/lesslinux/cryptpart/" + container_name + "' ")
			system("rm '/lesslinux/cryptpart/" + container_name + "' ")
			system("umount /lesslinux/cryptpart")
			return false, "Loop-Device konnte nicht erstellt werden."
		end
		system("mdev -s")
	end
	# fill the first 16 megabytes with random data
	unless system("dd if=/dev/urandom of=" + free_loop + " bs=1M count=16")
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht erstellt werden."
	end
	# establish cryptsetup on the container
	unless system("echo '" + passphrase + "' | cryptsetup luksFormat -c aes-cbc-essiv:sha256 -s 256 -q " + free_loop ) 
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht erstellt werden."
	end
	# open the encrypted container
	unless system("echo '" + passphrase + "' | cryptsetup luksOpen -q " + free_loop + " lesslinux_crypt" ) 
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht zugegriffen werden."
	else
		system("mdev -s")
	end
	# create filesystem on container
	unless system("mkfs.ext2 /dev/mapper/lesslinux_crypt")
		system("cryptsetup luksClose /dev/mapper/lesslinux_crypt")
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht formatiert werden."
	end
	# mount container
	system("mkdir /lesslinux/tmp_home")
	unless system("mount -t ext2 /dev/mapper/lesslinux_crypt /lesslinux/tmp_home")
		system("cryptsetup luksClose /dev/mapper/lesslinux_crypt")
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht gemountet werden."
	end
	# copy/synchronize /home to the container
	unless system("rsync -avHP /home/ /lesslinux/tmp_home/")
		system("umount /lesslinux/tmp_home/")
		system("cryptsetup luksClose /dev/mapper/lesslinux_crypt")
		system("losetup -d " + free_loop)
		system("rm '/lesslinux/cryptpart/" + container_name + "' ")
		system("umount /lesslinux/cryptpart")
		return false, "Loop-Device konnte nicht gemountet werden."
	end
	# create overlay for passwords etc.
	system("mkdir -m 0755 /lesslinux/tmp_home/.overlay")
	# unmount the container
	system("umount /lesslinux/tmp_home/")
	# loosen cryptsetup
	system("cryptsetup luksClose /dev/mapper/lesslinux_crypt")
	# deattach the loop device
	system("losetup -d " + free_loop)
	system("umount /lesslinux/cryptpart")
	return true, "Erstellung des Containers und Kopieren der Daten erfolgreich"
end

def create_crypt_tab (parent)
	
	partitions, containers, mnt_parts = find_containers
	
	fat_font = Pango::FontDescription.new("Sans Bold 12")
	p_swin = Gtk::ScrolledWindow.new
	p_swin.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)

	pt_box = Gtk::VBox.new(true, 0)
	ptable = Gtk::Table.new(6,    8,    false)

	pt_options = Gtk::EXPAND

	cont_label = Gtk::Label.new("Verschlüsselten Container verwenden")
	cont_label.modify_font(fat_font)
	creat_label = Gtk::Label.new("Verschlüsselten Container erstellen")
	creat_label.modify_font(fat_font)

	radio1 = Gtk::RadioButton.new("_Nein, mit LessLinux-Lesezeichen und -Einstellungen starten.")
	radio2 = Gtk::RadioButton.new(radio1, "_Ja, mit eigenen Einstellungen starten und Änderung speichern.")
	radio3 = Gtk::RadioButton.new(radio1, "Ja, eigene Einstellungen laden, aber Änderungen später _verwerfen.")
	combobox = Gtk::ComboBox.new
	use_crypt_button =  Gtk::Button.new("Container jetzt benutzen")
	if containers.size > 0 
		containers.each { |c|
			combobox.append_text( c[0] + ", Datei: " + c[1] )
		}
		combobox.active = 0
	else
		combobox.append_text( "Kein Container gefunden" )
		use_crypt_button.sensitive = false
		combobox.active = 0
	end
	
	
	rvbox = Gtk::VBox.new(false, 5)
	rvbox.pack_start(cont_label, false, true, 5)
	rvbox.pack_start(radio1, false, true, 1)
	rvbox.pack_start(radio2, false, true, 1)
	rvbox.pack_start(radio3, false, true, 1)

	chbox = Gtk::HBox.new(false, 5)
	chbox.pack_start(Gtk::Label.new("Auswahl des Containers:"), false, false, 1)
	create_crypt_button =  Gtk::Button.new("Neuer Container")
	chbox.pack_start(combobox, false, false, 1)
	chbox.pack_start(create_crypt_button, false, false, 1)
	rvbox.pack_start(chbox, false, true, 1)
	
	create_crypt_button.signal_connect( "clicked" ) { |w|
		retval, partition, container_name, size, pass, conf = create_container_dialog(partitions)
		pw_res, pw_mesg = pw_check(pass, conf)
		if pw_res == false
			if retval == true
				show_error_dialog(pw_mesg)
			end
		else
			# lock buttons before creating container
			create_crypt_button.sensitive = false
			use_crypt_button.sensitive = false
			create_success, create_message = create_crypt_container(partition, size, container_name, pass)
			create_crypt_button.sensitive = true
			if create_success == true
				choice_count = containers.size
				partitions, containers, mnt_parts = find_containers
				if choice_count > 0
					(choice_count - 1).downto(0) { |i| combobox.remove_text(i) }
				else
					combobox.remove_text(0)
				end
				containers.each { |c|
					combobox.append_text( c[0] + ", Datei: " + c[1] )
				}
				combobox.active = 0
				if containers.size > 0 
					combobox.sensitive = true
					radio2.sensitive = true
					radio3.sensitive = true
					use_crypt_button.sensitive = true
				end
			else
				show_error_dialog(create_message)
			end
		end
	}
	
	use_crypt_button.signal_connect( "clicked" ) {|w| 
		# signal, message = mount_crypt_container(containers, mnt_parts, radio1, radio2, radio3, combobox)
		# signal, message = mount_crypt_container([ [ "hulla.llc", "/dev/sdx1"] ] , mnt_parts, radio1, radio2, radio3, combobox)

		retval, passwd = mount_crypt_dialog
		puts retval.to_s + ", " + passwd.to_s
		if retval == true
			mount_success, mount_message = mount_crypt_container(containers, mnt_parts, 
				radio1, radio2, radio3, combobox, passwd)
			# mount_success, mount_message = mount_crypt_container([ [ "hulla.llc", "/dev/sdx1"] ], mnt_parts, 
			#	radio1, radio2, radio3, combobox, passwd)
			if  mount_success == true
				combobox.sensitive = false
				radio1.sensitive = false
				radio2.sensitive = false
				radio3.sensitive = false
				use_crypt_button.sensitive = false
				use_crypt_button.label = "Container bereits aktiv"
				create_crypt_button.sensitive = false
				puts mount_message
			else
				puts mount_message
				show_error_dialog(mount_message)
			end
		end
		
	}
	
	if containers.size < 1 
		combobox.sensitive = false
		radio2.sensitive = false
		radio3.sensitive = false
		use_crypt_button.sensitive = false
	end
	if partitions.size < 1
		create_crypt_button.sensitive = false
	end
	
	crypt_button_box = Gtk::HBox.new(false, 5)
	crypt_button_box.pack_start(use_crypt_button, false, true, 5)
	rvbox.pack_start(crypt_button_box, false, true, 5)
	# rvbox.pack_start(creat_label, false, true, 5)
	# sbox = Gtk::HBox.new(false, 5)
	# sscale = Gtk::HScale.new(64, 2048, 64)
	# sbox.pack_start(Gtk::Label.new("Größe des Containers:"), false, true, 1)
	# sbox.pack_start(sscale, true, true, 1)
	# rvbox.pack_start(sbox, false, true, 1)

	#rvbox.pack_start(radio4, false, true, 1)

	pt_box.pack_start(ptable, false, false, 0)
	p_swin.add_with_viewport(rvbox)
	return p_swin, Gtk::Label.new("Speicherung")
end

def create_passwd_tab
	# scrolled window for the tab users/password
	swin       = Gtk::ScrolledWindow.new
	swin.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
	table_box = Gtk::VBox.new(false, 0)
	table = Gtk::Table.new(4,    2,    false)

	t_options = Gtk::EXPAND|Gtk::FILL
	t_name = Gtk::Entry.new
	initial_font = Pango::FontDescription.new("Sans Bold 12")
	
	user_label = Gtk::Label.new("Nutzer-Passwort")
	user_label.modify_font(initial_font)
	root_label = Gtk::Label.new("Administrator-Passwort")
	root_label.modify_font(initial_font)
	user_desc = Gtk::Label.new("Sie benötigen das Nutzer-Passwort, wenn Sie im laufenden System Laufwerke " + 
		"ein- oder aushängen wollen. Wenn Sie auf interne Laufwerke oder USB-Sticks zugreifen wollen, " + 
		"vergeben Sie jetzt ein Passwort.")
	user_desc.wrap = true
	user_button = Gtk::Button.new("Nutzer-Passwort setzen")
	user_button.signal_connect( "clicked" ) {|w| change_password("surfer") }
	user_button_box = Gtk::VBox.new(true, 0)
	user_button_box.pack_start(user_button, false, false, 0)
	root_desc = Gtk::Label.new("Sie benötigen ein Administrator-Passwort nur, wenn Sie weitergehende " + 
		"Wartungsarbeiten durchführen wollen, oder wenn der Support Sie darum bittet, ein Admin-Passwort zu setzen.")
	root_desc.wrap = true
	root_button =  Gtk::Button.new("Admin-Passwort setzen")
	root_button.signal_connect( "clicked" ) {|w| change_password("root") }
	root_button_box = Gtk::VBox.new(true, 0)
	root_button_box.pack_start(root_button, false, false, 0)
	table.attach(user_label,  0,  2,  0,  1, Gtk::EXPAND, Gtk::SHRINK, 5,    25)
	table.attach(user_desc,  0,  1,  1,  2, Gtk::EXPAND|Gtk::FILL, Gtk::SHRINK, 5,    5)
	user_desc.set_alignment(0, 0.1 )
	table.attach(user_button_box,  1,  2,  1,  2, Gtk::SHRINK, Gtk::SHRINK, 5,    5)
	table.attach(root_label,  0,  2,  2,  3, Gtk::EXPAND, Gtk::SHRINK, 5,    25)
	table.attach(root_desc,  0,  1,  3,  4, Gtk::EXPAND|Gtk::FILL, Gtk::SHRINK, 5,    5)
	root_desc.set_alignment(0, 0.1 )
	table.attach(root_button_box,  1,  2,  3,  4, Gtk::SHRINK, Gtk::SHRINK, 5,    5)
	table_box.pack_start(table, false, false, 0)
	swin.add_with_viewport(table_box)
	return swin, Gtk::Label.new("Nutzer")
end

def adjust_buttons (notebook, active_tab, cont_button, back_button)
	puts '-> Tab switched ' + active_tab.to_s
	# puts notebook.n_pages
	if (active_tab < notebook.n_pages - 1)
		cont_button.set_label("Weiter")
	else
		cont_button.set_label("Beenden")
	end
	if (active_tab < 1)
		back_button.sensitive = false
	else
		back_button.sensitive = true
	end
end

def switch_tabs (notebook, clicked, continue, back)
	if notebook.page == notebook.n_pages - 1 && clicked == continue
		cleanup
		Gtk.main_quit
	end
	if clicked == back
		notebook.prev_page
	elsif clicked == continue	
		notebook.next_page
	end
	if (notebook.page > notebook.n_pages - 2)
		continue.set_label("Beenden")
	else
		continue.set_label("Weiter")
	end
	if (notebook.page == 0)
		back.sensitive = false
	else
		back.sensitive = true
	end
	
end

def show_error_dialog(message)
	dialog = Gtk::MessageDialog.new($main_application_window, 
			Gtk::Dialog::MODAL,
			Gtk::MessageDialog::ERROR,
			Gtk::MessageDialog::BUTTONS_CLOSE,
			message)
	dialog.run
	dialog.destroy
end

def run_installation(target, format, write_mbr)
	disk_device, disk_name, partitions = get_drive_details(target)
	# disk_device = "/dev/null" if disk_device.nil? # for testing purposes
	
	# try to mount
	system("mkdir -p /lesslinux/install")
	message = []
	unless system("mount -t vfat -o ro " + target + " /lesslinux/install")
		message.push("Mount fehlgeschlagen: Die Installation ist wahrscheinlich nur möglich, wenn die Zielpartition auch formatiert wird.")
	else
		# check for existing files
		check_files = [ "lesslinux", "syslinux.cfg" ]
		traces_found = nil
		Dir.entries("/lesslinux/install").each { |e|
			check_files.each { |f| traces_found = f if e == f }
		}
		unless traces_found.nil?
			message.push("Es wurden Dateien gefunden, die auf eine ältere Installation von LessLinux hinweisen.")
		end
		system("umount /lesslinux/install")
	end
	if partitions.size > 1
		message.push("Der USB-Stick oder die Festplatte " + disk_device.to_s + " ist in mehrere Partitionen aufgeteilt. Der Start anderer dort bootfähig installierter Betriebssysteme ist nach der Installation eventuell nicht mehr möglich." )
	end
	
	# FIXME: check for free space
	# FIXME: check if target and source are identical
	dialog = Gtk::Dialog.new("Installation",
		$main_application_window,
		Gtk::Dialog::MODAL,
		[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT],
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_ACCEPT])
	install_intro = Gtk::Label.new("Bitte prüfen Sie die Installationseinstellungen:")
	dialog.vbox.add(install_intro)	
	dialog.vbox.add(Gtk::HSeparator.new)
	message.each { |m|
		msg_label = Gtk::Label.new(" * " + m)
		msg_label.wrap = true
		dialog.vbox.add(msg_label)
	}
	dialog.vbox.add(Gtk::HSeparator.new)
	install_label = Gtk::Label.new("Wollen Sie LessLinux auf die Partition " + target.to_s + " auf dem Laufwerk " + disk_name.to_s + " installieren?")
	install_label.wrap = true
	dialog.vbox.add(install_label)	
	dialog.show_all
	retval = false
	if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
		retval = true
	end
	dialog.destroy
	
	if retval == true
		# FIXME: Check if something went wrong!
		# Format now
		if format == true 
			system("mkfs.msdos -F 32 " + target.to_s)
		end
		# find out if install source is still mounted
		unless system("mount | awk '{print $1}' | grep ` cat /var/run/lesslinux/install_source ` ")
			system("mount ` cat /var/run/lesslinux/install_source ` /lesslinux/cdrom")
		end
		# mount the target
		system("mkdir /lesslinux/install_target")
		system("mount " + target + " /lesslinux/install_target")
		# copy files
		[ "syslinux.cfg", "boot", "lesslinux"].each { |e|
			system("tar -C /lesslinux/cdrom -cf - " + e + " | tar -C  /lesslinux/install_target -xvf - ")
		}
		system("umount /lesslinux/install_target")
		system("umount /lesslinux/cdrom")
		# write MBR
		if write_mbr == true
			system("cat /usr/share/syslinux/mbr.bin > " + disk_device )
		end
		# set partition active
		system("parted " + disk_device + " set " + target.split(disk_device)[1] + " boot on ")
		# write bootloader
		system("syslinux " + target)
		return true, nil
	end
	return false, nil
end

def create_install_tab
	# partitions, containers, mnt_parts = find_containers
	usb_drives = get_usbdrives_lshw
	usable_partitions = []
	layout_vbox = Gtk::VBox.new(false, 5)
	fat_font = Pango::FontDescription.new("Sans Bold 12")
	heading = Gtk::Label.new("Installation")
	heading.modify_font(fat_font)
	desc = Gtk::Label.new("Installieren Sie LessLinux auf USB-Stick oder Festplatte, um das System immer am Schlüsselbund dabeizuhaben, einfacher Updates vornehmen und schneller starten zu können.")
	desc.wrap = true
	desc.set_alignment(0, 0.1 )
	desc.width_request = 550
	layout_vbox.pack_start(heading, false, false, 5)
	layout_vbox.pack_start(desc, false, false, 5)
	lwlabel = Gtk::Label.new("Laufwerk:")
	lwcombo = Gtk::ComboBox.new
	lwcombo.width_request = 360
	partbutton = Gtk::Button.new("Neue Partition")
	if usb_drives.size < 1
		partbutton.sensitive = false
	end
	lwhbox = Gtk::HBox.new(false, 2)
	lwhbox.pack_start(lwlabel, false, false,2)
	lwhbox.pack_start(lwcombo, false, false, 2)
	lwhbox.pack_start(partbutton, false, false, 2)
	layout_vbox.pack_start(lwhbox, false, false, 2)
	check_boot = Gtk::CheckButton.new("Bootloader schreiben")
	check_boot.active = true
	layout_vbox.pack_start(check_boot, false, false, 2)
	check_format = Gtk::CheckButton.new("Laufwerk formatieren")
	layout_vbox.pack_start(check_format, false, false, 2)
	install_button = Gtk::Button.new("Jetzt installieren")
	install_button_box = lwhbox = Gtk::HBox.new(false, 2)
	install_button_box.pack_start(install_button, false, false, 2)
	layout_vbox.pack_start(install_button_box, false, false, 2)
	
	cbrows = 0
	usb_drives.each { |key, val| 
		val.partitions.each { |part, det|
			if part[8..-1].to_i < 5 && det.mounted == false
				lwcombo.append_text( part + " " + (det.size / 1024 / 1024).to_i.to_s + 
					"MB (" + val.vendor + " " + val.product + " " + 
					((val.size.to_f / 1000000000.0) + 0.5).to_i.to_s + "GB)"  )
				usable_partitions.push(part)
				cbrows += 1
			end
		}
		lwcombo.active = 0
	}

	if usable_partitions.size < 1
		lwcombo.append_text( "Keine FAT-Partition gefunden" )
		install_button.sensitive = false
		lwcombo.active = 0
		lwcombo.sensitive = false
		check_boot.sensitive = false
		check_format.sensitive = false
	end
	partbutton.signal_connect( "clicked" ) {
		retval = create_partition_dialog(usb_drives)
		if retval == true
			(cbrows - 1).downto(0) { |i| lwcombo.remove_text(i) }
			cbrows = 0
			# REALLY IMPORTANT!
			usable_partitions = []
			usb_drives = get_usbdrives_lshw
			# FIXME: No copy and paste please
			usb_drives.each { |key, val| 
				val.partitions.each { |part, det|
					if part[8..-1].to_i < 5 && det.mounted == false
						lwcombo.append_text( part + " " + (det.size / 1024 / 1024).to_i.to_s + 
							"MB (" + val.vendor + " " + val.product + " " + 
							((val.size.to_f / 1000000000.0) + 0.5).to_i.to_s + "GB)"  )
						usable_partitions.push(part)
						cbrows += 1
					end
				}
				lwcombo.active = 0
				lwcombo.sensitive = true if cbrows > 0
			}
		end
	}
	install_button.signal_connect( "clicked" ) {|w|
		selected_partition = usable_partitions[lwcombo.active.to_i]
		#retval = run_installation(selected_partition, false)
		retval = run_installation(selected_partition, check_format.active?, check_boot.active?)
	}
	return layout_vbox, Gtk::Label.new("Installation")
end

def create_partition_dialog(drives)
	fat_font = Pango::FontDescription.new("Sans Bold 12")
	heading = Gtk::Label.new("USB-Laufwerk partitionieren")
	heading.modify_font(fat_font)
	desc = Gtk::Label.new("Teilen Sie ein USB-Laufwerk in zwei separate Abschnitte auf. Installieren Sie anschließend LessLinux (und ggf. den Datentresor) auf die zweite Partition. Die zweite Partition ist in der Regel unter Windows nicht sichtbar. Die erste Partition kann zum Datentausch verwendet werden. Achtung: Alle Daten werden gelöscht!")
	desc.width_request = 420
	desc.wrap = true
	lwlabel = Gtk::Label.new("Laufwerk:")
	lwcombo = Gtk::ComboBox.new
	lwcombo.width_request = 360
	lwhbox = Gtk::HBox.new(false, 2)
	lwhbox.pack_start(lwlabel, false, false,2)
	lwhbox.pack_start(lwcombo, false, false, 2)
	sizehbox = Gtk::HBox.new(false, 2)
	sizedesc = Gtk::Label.new("Größe der zweiten Partition (MB):")
	sizespin = Gtk::SpinButton.new(512.0, 200000.0, 1.0)
	sizespin.value = 1024
	sizehbox.pack_start(sizedesc, false, false,2)
	sizehbox.pack_start(sizespin, false, false,2)
	drive_devices = Array.new
	drives.each { |dev, det|
		mounted = false
		det.partitions.each { |part, pdet|
			mounted = true if pdet.mounted == true
		}
		if mounted == false && (det.size / 1000000).to_i > 998 
			lwcombo.append_text( dev + " (" + det.vendor + " " + det.product + ") " + (det.size / 1024 / 1024).to_i.to_s + "MB" )
			drive_devices.push(dev)
		end
	}
	if drive_devices.size < 1
		lwcombo.append_text("Keine nutzbaren Laufwerke gefunden")
		lwcombo.sensitive = false
	end
	lwcombo.active = 0
	dialog = Gtk::Dialog.new("USB-Laufwerk partitionieren",
		$main_application_window,
		Gtk::Dialog::MODAL,
		[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT],
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_ACCEPT])
	dialog.vbox.add(heading)
	dialog.vbox.add(desc)
	dialog.vbox.add(lwhbox)
	dialog.vbox.add(sizehbox)
	dialog.show_all
	retval = false
	if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT && drive_devices.size > 0
		retval = change_partitioning(drive_devices[lwcombo.active.to_i], sizespin.value.to_i)
	end
	dialog.destroy
	return retval
end

def change_partitioning(device, size)
	# First query details about the drive, we are just interested in he second line
	lcount = 0
	# use kB as units
	partsize = size * 1024
	drivesize = 0
	IO.popen("parted " + device + " unit kB print") { |line|
		while line.gets
			if lcount == 1
				drivesize = $_.split[2].to_i
				puts drivesize.to_s
			end
			lcount += 1
		end
	}
	# do some plausability checks 
	# partitioning just makes sense if the first partition has at least 128MB free space
	# FIXME: Currently fails silently!
	return false if (partsize + 131072 > drivesize)
	# delete all existing partitions
	15.downto(0) { |p| system("parted -s " + device + " rm " + p.to_s) }
	# calculate start and end of the second partition
	# leave some space at the end (other tools might use it to adust to cylinder boundaries)
	# this costs approximately 6MB in total
	part_two_end = drivesize - 4096
	part_two_start = part_two_end - partsize
	part_one_end = part_two_start - 2048
	part_one_start = 63
	# Now write the new partitioning including filesystems
	system("parted -s " + device + " unit kB mkpartfs primary fat32 " + part_one_start.to_s + " " + part_one_end.to_s)
	system("parted -s " + device + " unit kB mkpartfs primary fat32 " + part_two_start.to_s + " " + part_two_end.to_s)
	return true
end

def create_monitor_tab 
	
	fat_font = Pango::FontDescription.new("Sans Bold 12")
	p_swin = Gtk::ScrolledWindow.new
	p_swin.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
	head = Gtk::Label.new("Monitor und Grafikkarte")
	head.modify_font(fat_font)
	drivers = [ "Xorg (automatische Auswahl)", "VESA (langsam, funktioniert immer)" ]
	resolutions = [ 
		"1920x1200",  
		"1920x1080", 
		"1680x1050", 
		"1600x1200", 
		"1440x900", 
		"1400x1050", 
		"1366x768", 
		"1280x1024", 
		"1280x800", 
		"1152x864",
		"1024x768", 
		"1024x600", 
		"800x600", 
		"800x480", 
		"640x480" ]
	monitors = [ "automatisch erkennen", "Flachbildschirm/TFT (60Hz)", "Röhrenmonitor (75Hz)" ]
	monitor_modes = [ "AUTO", "TFT", "CRT" ]
	layout_table = Gtk::Table.new(5, 2, false)
	driver_combo = Gtk::ComboBox.new
	drivers.each { |i| driver_combo.append_text(i) }
	res_combo = Gtk::ComboBox.new
	resolutions.each { |i| res_combo.append_text(i) }
	res_combo.active = 9
	mon_combo = Gtk::ComboBox.new
	monitors.each { |i| mon_combo.append_text(i) }
	mon_combo.active = 0
	button_box = Gtk::HBox.new(true, 5)
	test_button = Gtk::Button.new("Test")
	save_button = Gtk::Button.new("Speichern")
	
	if File.exists?("/var/run/lesslinux/xconfgui_xorg")
		driver_combo.active = 0
	else
		driver_combo.active = 1
		test_button.sensitive = false
	end
	
	
	save_button.signal_connect("clicked") { |w|
		outfile = File.new("/var/run/lesslinux/xorg_vars", "w")
		# outfile = File.new("/tmp/xorg_vars", "w")
		if driver_combo.active == 0
			outfile.write("driver=xorg\n")
		else
			outfile.write("driver=xvesa\n")
		end
		res = resolutions[res_combo.active]
		outfile.write("vesapref=" + res + "x24\n")
		outfile.write("screen=Screen_" + monitor_modes[mon_combo.active] + "_" + res + "\n")
		outfile.close
	}
	
	test_button.signal_connect("clicked") { |w|
		outfile = File.new("/var/run/lesslinux/xorg_vars.test", "w")
		# outfile = File.new("/tmp/xorg_vars", "w")
		if driver_combo.active == 0
			outfile.write("driver=xorg\n")
		else
			outfile.write("driver=xvesa\n")
		end
		res = resolutions[res_combo.active]
		outfile.write("vesapref=" + res + "\n")
		outfile.write("screen=Screen_" + monitor_modes[mon_combo.active] + "_" + res + "\n")
		outfile.close
		system("/bin/bash /usr/local/xconfgui/trigger_configtest")
	}
	
	button_box.pack_start(test_button, true, true, 5)
	button_box.pack_start(save_button, true, true, 5)
	layout_table.attach(head, 0, 2, 0, 1 , Gtk::EXPAND, Gtk::SHRINK, 5, 15)
	layout_table.attach(driver_combo,  1,  2,  1,  2, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
	layout_table.attach(Gtk::Label.new("Treiber:"), 0,  1,  1,  2, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
	layout_table.attach(res_combo,  1,  2,  2,  3, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
	layout_table.attach(Gtk::Label.new("Auflösung:"), 0,  1,  2,  3, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
	layout_table.attach(mon_combo,  1,  2,  3,  4, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
	layout_table.attach(Gtk::Label.new("Monitor:"), 0,  1,  3,  4, Gtk::SHRINK, Gtk::SHRINK, 5, 2)
	layout_table.attach(button_box, 0,  2,  4,  5, Gtk::FILL|Gtk::EXPAND, Gtk::SHRINK, 5, 2)
	p_swin.add_with_viewport(layout_table)
	
	return p_swin, Gtk::Label.new("Grafik")
end

def cleanup
	[ "root", "surfer" ].each { |i|
		File.unlink("/etc/lesslinux/" + i + ".hash") if File.exists?("/etc/lesslinux/" + i + ".hash")
	}
end

# define base layout

window = Gtk::Window.new(Gtk::Window::TOPLEVEL)
window.set_title  "xconfgui"
window.border_width = 10
window.set_size_request(600, 440)
window.signal_connect('delete_event') { 
	cleanup
	Gtk.main_quit 
}

# notebook and tabs

nb = Gtk::Notebook.new
cont =  Gtk::Button.new("Weiter")
back =  Gtk::Button.new("Zurück")
cont.width_request = 80
back.width_request = 80

cont.signal_connect( "clicked" ) {|w| switch_tabs (nb, cont, cont, back) }
back.signal_connect( "clicked" ) {|w| switch_tabs (nb, back, cont, back) }
nb.signal_connect('switch-page') { |i,k,j| adjust_buttons (nb, j, cont, back) }
vbox = Gtk::VBox.new(false, 7)
bbox = Gtk::HBox.new(false, 0)

# Append the tabs

unless File.exists?("/var/run/lesslinux/usbfound")
	install_tab, install_label = create_install_tab
	nb.append_page(install_tab, install_label)
end

passwd_tab, passwd_label = create_passwd_tab 
nb.append_page(passwd_tab, passwd_label)
crypt_tab, crypt_label = create_crypt_tab(window)
nb.append_page(crypt_tab, crypt_label)

unless File.exists?("/var/run/lesslinux/xconfgui_skip_monitor")
	monitor_tab, monitor_label = create_monitor_tab
	nb.append_page(monitor_tab, monitor_label)
end 

vbox.pack_start(nb, true, true)
bbox.pack_end(cont, false, false, 0)
bbox.pack_end(back, false, false, 0)
vbox.pack_end(bbox, false, false, 0)

window.add(vbox)
window.show_all
Gtk.main
